"use strict";

var _isInteger = require("babel-runtime/core-js/number/is-integer");

var _isInteger2 = _interopRequireDefault(_isInteger);

var _getIterator2 = require("babel-runtime/core-js/get-iterator");

var _getIterator3 = _interopRequireDefault(_getIterator2);

var _map = require("babel-runtime/core-js/map");

var _map2 = _interopRequireDefault(_map);

var _defineProperty = require("babel-runtime/core-js/object/define-property");

var _defineProperty2 = _interopRequireDefault(_defineProperty);

var _symbol = require("babel-runtime/core-js/symbol");

var _symbol2 = _interopRequireDefault(_symbol);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Full credit: https://raw.githubusercontent.com/postgraphql/postgraphql/master/src/postgres/utils/sql.ts
var isSymbol = require("lodash/isSymbol");
var isNil = require("lodash/isNil");
var isPlainObject = require("lodash/isPlainObject");
var lodashIsFinite = require("lodash/isFinite");
var debug = require("debug")("pg-sql2");

var isDev = ["test", "development"].indexOf(process.env.NODE_ENV) >= 0;

function debugError(err) {
  debug(err);
  return err;
}

var $$trusted = (0, _symbol2.default)("trusted");
/*::
type SQLRawNode = {
  text: string,
  type: 'RAW',
  [Symbol]: true,
}
type SQLIdentifierNode = {
  names: Array<mixed>,
  type: 'IDENTIFIER',
  [Symbol]: true,
}
type SQLValueNode = {
  value: mixed,
  type: 'VALUE',
  [Symbol]: true,
}

type SQLNode = SQLRawNode | SQLValueNode | SQLIdentifierNode
*/

function makeTrustedNode /*:: <Node>*/(node /*: Node */) /*: Node */{
  (0, _defineProperty2.default)(node, $$trusted, {
    enumerable: false,
    configurable: false,
    value: true
  });
  return node;
}

function makeRawNode(text /*: string */) /*: SQLRawNode */{
  return makeTrustedNode({ type: "RAW", text });
}

function makeIdentifierNode(names /*: Array<mixed> */
) /*: SQLIdentifierNode */{
  return makeTrustedNode({ type: "IDENTIFIER", names });
}

function makeValueNode(value /*: mixed */) /*: SQLValueNode */{
  return makeTrustedNode({ type: "VALUE", value });
}

function ensureNonEmptyArray(array) {
  var allowZeroLength = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

  if (!Array.isArray(array)) {
    throw debugError(new Error("Expected array"));
  }
  if (array.length < 1 && !allowZeroLength) {
    throw debugError(new Error("Expected non-empty array"));
  }
  array.forEach(function (entry, idx) {
    if (entry == null) {
      throw debugError(new Error(`Array index ${idx} is ${String(entry)}`));
    }
  });
  return array;
}

function compile(sql /*: Array<SQLNode> */) {
  // Join this to generate the SQL query
  var sqlFragments = [];

  // Values hold the JavaScript values that are represented in the query
  // string by placeholders. They are eager because they were provided before
  // compile time.
  var values = [];

  // When we come accross a symbol in our identifier, we create a unique
  // alias for it that shouldn’t be in the users schema. This helps maintain
  // sanity when constructing large Sql queries with many aliases.
  var nextSymbolId = 0;
  var symbolToIdentifier = new _map2.default();

  var items = Array.isArray(sql) ? sql : [sql];

  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = (0, _getIterator3.default)(items), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var item = _step.value;

      if (!item[$$trusted]) {
        throw new Error(`Expecte sql item, instead got '${String(item)}'.`);
      }
      switch (item.type) {
        case "RAW":
          sqlFragments.push(item.text);
          break;
        case "IDENTIFIER":
          if (item.names.length === 0) throw new Error("Identifier must have a name");

          sqlFragments.push(item.names.map(function (rawName) {
            if (typeof rawName === "string") {
              var _name /*: string */ = rawName;
              return escapeSqlIdentifier(_name);
            }
            if (!isSymbol(rawName)) {
              throw debugError(new Error(`Expected string or symbol, received '${String(rawName)}'`));
            }
            var name /*: Symbol */ = /*:: (*/rawName /*: any) */;

            // Get the correct identifier string for this symbol.
            var identifier = symbolToIdentifier.get(name);

            // If there is no identifier, create one and set it.
            if (!identifier) {
              identifier = `__local_${nextSymbolId++}__`;
              symbolToIdentifier.set(name, identifier);
            }

            // Return the identifier. Since we create it, we won’t have to
            // escape it because we know all of the characters are safe.
            return identifier;
          }).join("."));
          break;
        case "VALUE":
          values.push(item.value);
          sqlFragments.push(`$${values.length}`);
          break;
        default:
      }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  var text = sqlFragments.join("");
  return {
    text,
    values
  };
}

/**
 * A template string tag that creates a `Sql` query out of some strings and
 * some values. Use this to construct all PostgreSQL queries to avoid SQL
 * injection.
 *
 * Note that using this function, the user *must* specify if they are injecting
 * raw text. This makes a SQL injection vulnerability harder to create.
 */
function query(strings /*: mixed */) /*: Array<mixed> */{
  for (var _len = arguments.length, values = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    values[_key - 1] = arguments[_key];
  }

  if (!Array.isArray(strings)) {
    throw new Error("sql.query should be used as a template literal, not a function call!");
  }
  return strings.reduce(function (items, text, i) {
    if (typeof text !== "string") {
      throw new Error("sql.query should be used as a template literal, not a function call.");
    }
    if (!values[i]) {
      return items.concat(makeRawNode(text));
    } else {
      var _value = values[i];
      if (isDev) {
        // These errors don't give you additional safety, they just catch
        // mistakes earlier to aid debugging, so they're fine to disable in
        // production
        if (!Array.isArray(_value) && !isPlainObject(_value)) {
          if (typeof _value === "string") {
            throw new Error(`Raw string passed into SQL query: '${String(_value)}'.`);
          } else if (typeof _value === "number") {
            throw new Error(`Raw number passed into SQL query: '${String(_value)}'.`);
          } else {
            throw new Error(`Invalid raw value passed into SQL query: '${String(_value)}'.`);
          }
        }
      }
      return items.concat(makeRawNode(text), _value);
    }
  }, []);
}

/**
 * Creates a Sql item for some raw Sql text. Just plain ol‘ raw Sql. This
 * method is dangerous though because it involves no escaping, so proceed
 * with caution!
 */
var raw = function raw(text /*: mixed */) {
  return makeRawNode(String(text));
};

/**
 * Creates a Sql item for a Sql identifier. A Sql identifier is anything like
 * a table, schema, or column name. An identifier may also have a namespace,
 * thus why many names are accepted.
 */
var identifier = function identifier() {
  for (var _len2 = arguments.length, names = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    names[_key2] = arguments[_key2];
  }

  return (/*: Array<mixed> */makeIdentifierNode(ensureNonEmptyArray(names))
  );
};

/**
 * Creates a Sql item for a value that will be included in our final query.
 * This value will be added in a way which avoids Sql injection.
 */
var value = function value(val /*: mixed */) {
  return makeValueNode(val);
};

/**
 * If the value is simple will inline it into the query, otherwise will defer
 * to value.
 */
var literal = function literal(val /*: mixed */) {
  if (typeof val === "string" && val.match(/^[a-zA-Z0-9_-]*$/)) {
    return raw(`'${val}'`);
  } else if (typeof val === "number" && lodashIsFinite(val)) {
    if ((0, _isInteger2.default)(val)) {
      return raw(String(val));
    } else {
      return raw(`'${0 + val}'::float`);
    }
  } else if (typeof val === "boolean") {
    if (val) {
      return raw(`TRUE`);
    } else {
      return raw(`FALSE`);
    }
  } else if (isNil(val)) {
    return raw(`NULL`);
  } else {
    return makeValueNode(val);
  }
};

/**
 * Join some Sql items together seperated by a string. Useful when dealing
 * with lists of Sql items that doesn’t make sense as a Sql query.
 */
var join = function join(rawItems /*: mixed */) {
  var rawSeparator /*: mixed */ = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "";

  if (!Array.isArray(rawItems)) {
    throw new Error("Items to join must be an array");
  }
  var items = rawItems;
  if (typeof rawSeparator !== "string") {
    throw new Error("Invalid separator - must be a string");
  }
  var separator = rawSeparator;
  return ensureNonEmptyArray(items, true).reduce(function (currentItems, item, i) {
    if (i === 0 || !separator) {
      return currentItems.concat(item);
    } else {
      return currentItems.concat(makeRawNode(separator), item);
    }
  }, []);
};

// Copied from https://github.com/brianc/node-postgres/blob/860cccd53105f7bc32fed8b1de69805f0ecd12eb/lib/client.js#L285-L302
// Ported from PostgreSQL 9.2.4 source code in src/interfaces/libpq/fe-exec.c
function escapeSqlIdentifier(str) {
  var escaped = '"';

  for (var i = 0; i < str.length; i++) {
    var c = str[i];
    if (c === '"') {
      escaped += c + c;
    } else {
      escaped += c;
    }
  }

  escaped += '"';

  return escaped;
}

exports.query = query;
exports.fragment = query;
exports.raw = raw;
exports.identifier = identifier;
exports.value = value;
exports.literal = literal;
exports.join = join;
exports.compile = compile;
exports.null = literal(null);
exports.blank = query``;